"""Module containing BaseCallBackMixin implementation"""
import sys
import inspect


class ExtendedCallBackMixin:
    """Class implementing callback decorator.

    Attributes:
        N/A

    Methods:
        validate_pass_vars():
            Validates `pass_vars` constructor argument.
        validate_exception_classes():
            Validates `exception_classes` constructor argument.
        get_func_scope_vars():
            Gets requested decorated function's scope variables, specified in
            `pass_vars` attribute.
        set_tracer_profile(tracer):
            Sets custom tracer profile.
        tracer(frame, event, _):
            Represents tracer storing local variables from executed function.
    """

    def __init__(
        self,
        default_return=None,
        pass_vars=None,
        exception_classes=(Exception,),
        **kwargs,
    ):
        """Class constructor. Sets instance variables.

        Args:
            default_return (Any, optional): Result to be returned in case
                of error or no return. Defaults to None.
            pass_vars (list|tuple|set, optional): Variable names to be passed
                to `on_end` callback. Defaults to None.
            exception_classes (list|tuple|set): Exception classes to be caught.
                Defaults to (Exception,).

        Returns:
            ExtendedCallBackMixin: mixin instance
        Raises:
            N/A
        """
        super().__init__(**kwargs)
        self.default_return = default_return
        self.pass_vars = pass_vars
        self.exception_classes = exception_classes
        self.local_vars = {}

    def validate_pass_vars(self):
        """Validates `pass_vars` constructor argument.

        Executes following checks:
            1. Parameter `pass_vars` must be `list`, `tuple` or `set`.
            2. Variable in `pass_vars` must be of type `str`.

        Args:
            N/A
        Returns:
            None
        Raises:
            TypeError: Raised if `pass_vars` is not `list`, `tuple` or `set`.
            TypeError: Raised if one of `pass_vars` elements is not `str`.
        """
        if not self.pass_vars:
            return
        if not isinstance(self.pass_vars, (list, tuple, set)):
            raise TypeError("`pass_vars` must be `list`, `tuple` or `set`.")
        for var in self.pass_vars:
            if not isinstance(var, str):
                raise TypeError("`pass_vars` elements must be of type `str`.")

    def validate_exception_classes(self):
        """Validates `exception_classes` constructor argument.

        Executes following checks:
            1. Argument `exception_classes` must be `tuple`, `list` or `set`.
            2. Elements of `exception_classes` must subclass of `Exception`.

        Args:
            N/A
        Returns:
            None
        Raises:
            TypeError: `exception_classes` is not `tuple`, `list` or `set`
            TypeError: Raised if any element of `exception_classes` is not
                subclass of `Exception`.
        """
        if not isinstance(self.exception_classes, (list, tuple, set)):
            raise TypeError("`exception_classes` not `list`, `tuple` or `set`")
        for exception_class in self.exception_classes:
            if not inspect.isclass(exception_class):
                raise TypeError("Element of `exception_classes` not a class.")
            if not issubclass(exception_class, Exception):
                raise TypeError(
                    "`exception_classes` must be a subclass of `Exception`."
                )

    def get_func_scope_vars(self):
        """Gets requested decorated function's scope variables, specified in
        `pass_vars` attribute. Any missing requested scope variables will be
        replaced by `<not-found>` string literal.

        Args:
            N/A
        Returns:
            dict: Names and values map of requested decorated function's
                scope variables.
        Raises:
            N/A
        """
        func_scope_vars = {}
        if not self.pass_vars:
            return {}
        for var_name in self.pass_vars:
            func_scope_vars[var_name] = self.local_vars.get(
                var_name,
                "<not-found>",
            )
        self.local_vars = {}
        return func_scope_vars

    def set_tracer_profile(self, tracer):
        """Sets custom tracer to the sys profile.

        Args:
            tracer (Tracer): Tracer to be set.
        Returns:
            None
        Raises:
            N/A
        """
        sys.setprofile(tracer)

    def tracer(self, frame, event, _):
        """Represents tracer storing local variables from executed function.
        Upon function return, tracer saves function locals to `local_vars`.

        Args:
            frame (Frame): Frame to be traced.
            event (Event): Event for tracing.
        Returns:
            None
        Raises:
            N/A
        """
        if event == "return":
            self.local_vars = frame.f_locals.copy()



import pytest

from callpyback.mixins.extended import ExtendedCallBackMixin


class Test___init__:
    """Class containing tests for __init__ method"""

    def test_constructor_defaults(self):
        """Tests that constructor sets default values correctly."""
        # Mocks
        # Calls
        mixin = ExtendedCallBackMixin()
        # Assertions
        assert mixin
        assert isinstance(mixin, ExtendedCallBackMixin)
        assert mixin.default_return is None
        assert mixin.pass_vars is None
        assert mixin.exception_classes == (Exception,)
        assert mixin.local_vars == {}

    def test_constructor_supplied_values(self):
        """Tests that constructor sets supplied values correctly."""
        # Mocks
        default_return = "default_return"
        pass_vars = ["var1", "var2"]
        exception_classes = (KeyError, ValueError)
        # Calls
        mixin = ExtendedCallBackMixin(
            default_return=default_return,
            pass_vars=pass_vars,
            exception_classes=exception_classes,
        )
        # Assertions
        assert mixin
        assert isinstance(mixin, ExtendedCallBackMixin)
        assert mixin.default_return == "default_return"
        assert mixin.pass_vars == ["var1", "var2"]
        assert mixin.exception_classes == (KeyError, ValueError)
        assert mixin.local_vars == {}


def create_mixin_obj(
    default_return=None,
    pass_vars=None,
    exception_classes=(Exception,),
):
    """Construct instance of BaseCallBackMixin object from supplied
    or default values."""
    instance = ExtendedCallBackMixin(
        default_return=default_return,
        pass_vars=pass_vars,
        exception_classes=exception_classes,
    )
    return instance


class Test_validate_pass_vars:
    """Class containing tests for validate_pass_vars method"""

    def test_correct(self):
        """Tests that no errors are raised when pass_vars are correct."""
        # Mocks
        mixin_obj = create_mixin_obj(pass_vars=("var1", "var2"))
        # Calls
        mixin_obj.validate_pass_vars()
        # Assertions
        assert True

    def test_empty(self):
        """Tests that no errors are raised when pass_vars is empty."""
        # Mocks
        mixin_obj = create_mixin_obj()
        # Calls
        mixin_obj.validate_pass_vars()
        # Assertions
        assert True

    def test_bad_list_type_error(self):
        """Tests that error is raised when `pass_vars` is not a list,
        tuple or set."""
        # Mocks
        mixin_obj = create_mixin_obj(pass_vars="bad_type")
        # Calls
        # Assertions
        with pytest.raises(
            TypeError, match="`pass_vars` must be `list`, `tuple` or `set`."
        ):
            mixin_obj.validate_pass_vars()

    def test_bad_variable_type_error(self):
        """Tests that error is raised when one of `pass_vars` elements
        is not a string."""
        # Mocks
        mixin_obj = create_mixin_obj(pass_vars=("x", 1))
        # Calls
        # Assertions
        with pytest.raises(
            TypeError, match="`pass_vars` elements must be of type `str`."
        ):
            mixin_obj.validate_pass_vars()


class Test_validate_exception_classes:
    """Class containing tests for validate_exception_classes method"""

    def test_list_type_error(self):
        """Tests that error is raised when `exception_classes` is not a list,
        tuple or set."""
        # Mocks
        mixin_obj = create_mixin_obj(exception_classes="not_a_list")
        # Calls
        # Assertions
        with pytest.raises(
            TypeError,
            match="`exception_classes` not `list`, `tuple` or `set`",
        ):
            mixin_obj.validate_exception_classes()

    def test_not_a_class_error(self):
        """Tests that error is raised when one of the `exception_classes`
        elements is not a class."""
        # Mocks
        mixin_obj = create_mixin_obj(exception_classes=("some_string",))
        # Calls
        # Assertions
        with pytest.raises(
            TypeError,
            match="Element of `exception_classes` not a class.",
        ):
            mixin_obj.validate_exception_classes()

    def test_exception_type_error(self):
        """Tests that error is raised when one of the `exception_classes`
        elements is not a subclass of `Exception`."""

        # Mocks
        class DummyClass:
            pass

        mixin_obj = create_mixin_obj(exception_classes=(DummyClass,))
        # Calls
        # Assertions
        with pytest.raises(
            TypeError,
            match="`exception_classes` must be a subclass of `Exception`.",
        ):
            mixin_obj.validate_exception_classes()


class Test_get_func_scope_vars:
    """Test get_func_scope_vars method"""

    def test_undefined_pass_vars(self):
        """Tests that `get_func_scope_vars` returns empty dict if `pass_vars`
        are not defined."""
        # Mocks
        mixin_obj = create_mixin_obj()
        # Calls
        func_scope_vars = mixin_obj.get_func_scope_vars()
        # Assertions
        assert func_scope_vars == {}

    def test_defined_pass_vars(self):
        """Tests that `get_func_scope_vars` returns full dict if `pass_vars`
        are defined and found in function scope variables."""
        # Mocks
        mixin_obj = create_mixin_obj(pass_vars=("var1", "var2"))
        mixin_obj.local_vars = {"var1": "value1", "var3": "value3"}
        # Calls
        func_scope_vars = mixin_obj.get_func_scope_vars()
        # Assertions
        assert func_scope_vars == {"var1": "value1", "var2": "<not-found>"}


class Test_set_tracer_profile:
    """Test set_tracer_profile method."""

    def test_default_tracer(self):
        """Tests that setting tracer profile executes successfully."""
        # Mocks
        mixin_obj = create_mixin_obj()
        # Calls
        mixin_obj.set_tracer_profile(None)
        # Assertions
        assert True


class Test_tracer:
    """Test tracer method"""

    def test_return_event(self):
        """Tests that local scope variables are extracted and saved to
        the instance attribute when `return` event is triggered."""
        # Mocks
        mixin_obj = create_mixin_obj()

        class DummyFrame:
            def __init__(self, f_locals):
                self.f_locals = f_locals

        frame = DummyFrame({"x": "y"})
        assert mixin_obj.local_vars == {}
        # Calls
        mixin_obj.tracer(frame, "return", None)
        # Assertions
        assert mixin_obj.local_vars == {"x": "y"}

    def test_other_event(self):
        """Test that other events than `return` are ignored."""
        # Mocks
        mixin_obj = create_mixin_obj()

        class DummyFrame:
            def __init__(self, f_locals):
                self.f_locals = f_locals

        frame = DummyFrame({"x": "y"})
        assert mixin_obj.local_vars == {}
        # Calls
        mixin_obj.tracer(frame, "yield", None)
        # Assertions
        assert mixin_obj.local_vars == {}



import pytest
from unittest.mock import MagicMock

from callpyback.mixins.base import BaseCallBackMixin
from callpyback.utils import _default_callback, background_callpyback


class Test___init__:
    """Class containing tests for __init__ method"""

    def test_constructor_defaults(self):
        """Tests that constructor sets default values correctly."""
        # Mocks
        # Calls
        mixin = BaseCallBackMixin()
        # Assertions
        assert mixin
        assert isinstance(mixin, BaseCallBackMixin)
        assert mixin.on_call is _default_callback
        assert mixin.on_success is _default_callback
        assert mixin.on_failure is _default_callback
        assert mixin.on_end is _default_callback
        assert mixin.callbacks == (
            _default_callback,
            _default_callback,
            _default_callback,
            _default_callback,
        )

    def test_constructor_supplied_values(self):
        """Tests that constructor sets supplied values correctly."""
        # Mocks
        on_call = MagicMock()
        on_success = MagicMock()
        on_failure = MagicMock()
        on_end = MagicMock()
        # Calls
        mixin = BaseCallBackMixin(
            on_call=on_call,
            on_success=on_success,
            on_failure=on_failure,
            on_end=on_end,
        )
        # Assertions
        assert mixin
        assert isinstance(mixin, BaseCallBackMixin)
        assert mixin.on_call == on_call
        assert mixin.on_success == on_success
        assert mixin.on_failure == on_failure
        assert mixin.on_end == on_end
        assert mixin.callbacks == (
            on_call,
            on_success,
            on_failure,
            on_end,
        )
        on_call.assert_not_called()
        on_success.assert_not_called()
        on_failure.assert_not_called()
        on_end.assert_not_called()


def create_mixin_obj(
    on_call=MagicMock(),
    on_success=MagicMock(),
    on_failure=MagicMock(),
    on_end=MagicMock(),
):
    """Construct instance of BaseCallBackMixin object from supplied or
    default values."""
    instance = BaseCallBackMixin(
        on_call=on_call,
        on_success=on_success,
        on_failure=on_failure,
        on_end=on_end,
    )
    return instance


class Test_validate_callbacks:
    """Class containing tests for validate_callbacks method"""

    def test_correct(self):
        """Tests that no errors are raised when callbacks are correct."""
        # Mocks
        mixin_obj = create_mixin_obj(
            on_call=_default_callback,
            on_success=_default_callback,
            on_failure=_default_callback,
            on_end=_default_callback,
        )
        # Calls
        mixin_obj.validate_callbacks()
        # Assertions
        assert True

    def test_not_callable_error(self):
        """Tests that error is raised when one of the callbacks is not
        Callable."""
        # Mocks
        mixin_obj = create_mixin_obj(
            on_call="not_callable",
        )
        # Calls
        # Assertions
        with pytest.raises(TypeError, match="Callback must be a callable."):
            mixin_obj.validate_callbacks()

    def test_coroutine_error(self):
        """Tests that error is raised when one of the callbacks is coroutine"""

        # Mocks
        async def test_coroutine():
            pass

        mixin_obj = create_mixin_obj(
            on_call=test_coroutine,
        )
        # Calls
        # Assertions
        with pytest.raises(
            TypeError, match="Callback `test_coroutine` cannot be a coroutine."
        ):
            mixin_obj.validate_callbacks()


class Test_validate_callback_parameters:
    """Class containing tests for validate_callback_parameters method"""

    def test_correct(self):
        """Tests that no errors are raised when callback parameters are ok."""

        # Mocks
        def on_success(func_result, func_kwargs):
            pass

        mixin_obj = create_mixin_obj(on_success=on_success)

        # Calls
        mixin_obj.validate_callback_parameters(on_success, "on_success")
        # Assertions
        assert True

    def test_bad_parameters(self):
        """Tests that error is raised when one of the callbacks accepts wrong
        combination of parameters."""

        # Mocks
        def on_success(x, y):
            pass

        mixin_obj = create_mixin_obj(
            on_success=on_success,
        )
        expected_error_message = (
            "Signature of callback `on_success` is invalid.\n"
            "Expected: No parameter or combination of: "
            "func,func_kwargs,func_result.\n"
            "Found: x,y."
        )
        # Calls
        # Assertions
        with pytest.raises(AssertionError, match=expected_error_message):
            mixin_obj.validate_callback_parameters(on_success, "on_success")


class Test_run_callback_func:
    """Test run_callback_func method"""

    def test_blocking_callback(self):
        """Tests that blocking callback executes successfully synchronously."""
        # Mocks
        mixin_obj = create_mixin_obj()
        func = MagicMock()
        func_kwargs = {"var": "value"}
        # Calls
        mixin_obj.run_callback_func(func, func_kwargs)
        # Assertions
        func.assert_called_once_with(**func_kwargs)

    def test_background_callback(self):
        """Tests that background callback (marked by @background_callpyback)
        executes successfully on a new thread (in a background)."""
        # Mocks
        mixin_obj = create_mixin_obj()
        mock = MagicMock()
        func_kwargs = {"var": "value"}

        @background_callpyback
        def func(*_, **__):
            mock()

        # Calls
        assert hasattr(func, "__background_callpyback__")
        mixin_obj.run_callback_func(func, func_kwargs)
        # Assertions
        mock.assert_called_once()


class Test_run_on_call_func:
    """Test run_on_call_func method"""

    def test_basic(self):
        """Tests that `on_call` callback function gets its kwargs generated
        and is called."""
        # Mocks
        on_call = MagicMock()
        func = MagicMock()
        mixin_obj = create_mixin_obj(on_call=on_call)
        func_kwargs = {"var": "value"}
        mixin_obj.get_callback_kwargs = MagicMock(
            return_value={"var1": "value1"},
        )
        mixin_obj.run_callback_func = MagicMock()
        # Calls
        mixin_obj.run_on_call_func(func, func_kwargs)
        # Assertions
        mixin_obj.get_callback_kwargs.assert_called_once_with(
            callback=on_call, func=func, func_kwargs=func_kwargs
        )
        mixin_obj.run_callback_func.assert_called_once_with(
            on_call,
            {"var1": "value1"},
        )


class Test_run_on_success_func:
    """Test run_on_success method"""

    def test_basic(self):
        """Tests that `on_success` callback function gets its kwargs generated
        and is called."""
        # Mocks
        on_success = MagicMock()
        mixin_obj = create_mixin_obj(on_success=on_success)
        func_kwargs = {"var": "value"}
        func_result = -1
        mixin_obj.get_callback_kwargs = MagicMock(
            return_value={"var1": "value1"},
        )
        mixin_obj.run_callback_func = MagicMock()
        func = MagicMock()
        # Calls
        mixin_obj.run_on_success_func(func, func_result, func_kwargs)
        # Assertions
        mixin_obj.get_callback_kwargs.assert_called_once_with(
            callback=on_success,
            func=func,
            func_result=func_result,
            func_kwargs=func_kwargs,
        )
        mixin_obj.run_callback_func.assert_called_once_with(
            on_success, {"var1": "value1"}
        )


class Test_run_on_failure_func:
    """Test run_on_failure method"""

    def test_basic(self):
        """Tests that `on_failure` callback function gets its kwargs generated
        and is called."""
        # Mocks
        on_failure = MagicMock()
        mixin_obj = create_mixin_obj(on_failure=on_failure)
        func_kwargs = {"var": "value"}
        func_exception = Exception("some error")
        mixin_obj.get_callback_kwargs = MagicMock(
            return_value={"var1": "value1"},
        )
        mixin_obj.run_callback_func = MagicMock()
        func = MagicMock()
        # Calls
        mixin_obj.run_on_failure_func(func, func_exception, func_kwargs)
        # Assertions
        mixin_obj.get_callback_kwargs.assert_called_once_with(
            callback=on_failure,
            func=func,
            func_exception=func_exception,
            func_kwargs=func_kwargs,
        )
        mixin_obj.run_callback_func.assert_called_once_with(
            on_failure, {"var1": "value1"}
        )


class Test_run_on_end_func:
    """Test run_on_end method"""

    def test_basic(self):
        """Tests that `on_end` callback function gets its kwargs generated
        and is called."""
        # Mocks
        on_end = MagicMock()
        mixin_obj = create_mixin_obj(on_end=on_end)
        func_kwargs = {"var": "value"}
        func_result = -1
        func_scope_vars = {"var1": "value1"}
        func_exception = Exception("some error")
        mixin_obj.get_callback_kwargs = MagicMock(
            return_value={"var1": "value1"},
        )
        mixin_obj.run_callback_func = MagicMock()
        func = MagicMock()
        # Calls
        mixin_obj.run_on_end_func(
            func,
            func_result,
            func_exception,
            func_kwargs,
            func_scope_vars,
        )
        # Assertions
        mixin_obj.get_callback_kwargs.assert_called_once_with(
            callback=on_end,
            func=func,
            func_result=func_result,
            func_exception=func_exception,
            func_kwargs=func_kwargs,
            func_scope_vars=func_scope_vars,
        )
        mixin_obj.run_callback_func.assert_called_once_with(
            on_end,
            {"var1": "value1"},
        )


class Test_get_callback_kwargs:
    """Test get_callback_kwargs method"""

    def test_no_kwargs(self):
        """Tests that `get_callback_kwargs` returns empty dict if no
        parameters are defined in the signature of callback function."""

        # Mocks
        def on_success():
            pass

        func_kwargs = {"var1": "key1"}
        func_result = -1
        mixin_obj = create_mixin_obj(on_success=on_success)
        # Calls
        on_call_kwargs = mixin_obj.get_callback_kwargs(
            callback=on_success,
            func_kwargs=func_kwargs,
            func_result=func_result,
        )
        # Assertions
        assert on_call_kwargs == {}

    def test_all_kwargs(self):
        """Tests that `get_callback_kwargs` returns all parameters that are
        defined in the signature of callback function."""

        # Mocks
        def on_success(func_result, func_kwargs):
            pass

        func_kwargs = {"var1": "key1"}
        func_result = -1
        mixin_obj = create_mixin_obj(on_success=on_success)
        # Calls
        on_call_kwargs = mixin_obj.get_callback_kwargs(
            callback=on_success,
            func_result=func_result,
            func_kwargs=func_kwargs,
        )
        # Assertions
        assert on_call_kwargs == {
            "func_kwargs": func_kwargs,
            "func_result": func_result,
        }



from callpyback.utils import (
    background_callpyback,
    _default_callback,
    args_to_kwargs,
)


class Test_background_callback:
    """Class containing tests for background_callback function"""

    def test_attribute_set(self):
        """Test whether the background_callback attribute is set
        to the function."""

        # Mocks
        @background_callpyback
        def function_x():
            pass

        # Calls
        function_x()
        # Assertions
        assert hasattr(function_x, "__background_callpyback__")
        assert function_x.__background_callpyback__


class Test__default_callback:
    """Class containing tests for __default_callback function"""

    def test_call(self):
        """Test that _default_callback function executes without errors."""
        # Mocks
        # Calls
        _default_callback()
        # Assertions
        assert True


class Test_args_to_kwargs:
    """Class containing tests for args_to_kwargs function"""

    def test_basic(self):
        """Test that args_to_kwargs function returns a combined kwargs."""
        # Mocks
        args = (1, 2, 3)
        kwargs = {"d": "val1", "e": "val2"}

        def func(a, b, c, d, e):
            pass

        # Calls
        result = args_to_kwargs(func, args, kwargs)
        # Assertions
        assert result == {"a": 1, "b": 2, "c": 3, "d": "val1", "e": "val2"}



"""Utility functions"""
import inspect


def background_callpyback(func):
    """Decorator to mark a callback function as a background task.
    Marking is done by assigning an attribute `background_callpyback`
    to the decorated function. This will cause the callback to be executed
    in a separate thread without blocking.

    Args:
        func (Callable): Decorated function to be marked as a background task.

    Returns:
        Callable: Decorated function marked as a background task.
    Raises:
        N/A
    """
    func.__background_callpyback__ = True
    return func


def _default_callback():
    """Default callback function to be used when callback is not provided."""


def args_to_kwargs(func, func_args, func_kwargs):
    """Transforms `args` and `kwargs` to `all_func_kwargs`.

    Args:
        func(Callable): Decorated function to be executed amongst callbacks.
        func_args(tuple): Arguments for the decorated function.
        func_kwargs(dict): Keyword arguments for the decorated function.
    Returns:
        dict: args and kwargs combined into kwargs for decorated function.
    Raises:
        N/A
    """
    all_kwargs = {}
    func_parameters = inspect.signature(func).parameters
    for param, arg in zip(func_parameters.keys(), func_args):
        all_kwargs[param] = arg
    for kwarg_name, kwarg_value in func_kwargs.items():
        all_kwargs[kwarg_name] = kwarg_value
    return all_kwargs



"""Module containing CallPyBack implementation"""

from callpyback.utils import _default_callback, args_to_kwargs
from callpyback.mixins.base import BaseCallBackMixin
from callpyback.mixins.extended import ExtendedCallBackMixin


class CallPyBack(BaseCallBackMixin, ExtendedCallBackMixin):
    """Class implementing callback decorator.

    Attributes:
        N/A

    Methods:
        validate_arguments():
            Runs validation functions for constructor arguments.
        validate_across_mixins():
            Performs cross-mixin validation, which cannot be tested separately.
        __call__(func):
            Invoked on decorator instance call.
        main(func, *args, **kwargs):
            Contains the actual callback logic.
    """

    def __init__(self, **kwargs):
        """Class constructor. Sets instance variables.

        Args:
            on_call (Callable, optional): Called before function execution.
                Defaults to DEFAULT_ON_CALL_LAMBDA.
            on_success (Callable, optional): Called after successfull run.
                Defaults to DEFAULT_ON_SUCCESS_LAMBDA.
            on_failure (Callable, optional): Called after run with errors.
                Defaults to DEFAULT_ON_FAILURE_LAMBDA.
            on_end (Callable, optional): Called after every execution.
                Defaults to DEFAULT_ON_END_LAMBDA.
            default_return (Any, optional): Result to be returned in
                case of error or no return. Defaults to None.
            pass_vars (list|tuple|set, optional): Variable names to be passed
                to `on_end` callback. Defaults to None.
            exception_classes (list|tuple|set): Exception classes to be caught.
                Defaults to (Exception,).

        Returns:
            CallPyBack: decorator instance
        Raises:
            N/A
        """
        super().__init__(**kwargs)

    def validate_arguments(self):
        """Runs validation functions for constructor arguments.

        Args:
            None
        Returns:
            N/A
        """
        self.validate_callbacks()
        self.validate_pass_vars()
        self.validate_exception_classes()
        self.validate_across_mixins()

    def validate_across_mixins(self):
        """Performs cross-mixin validation, which cannot be tested separately.

        Executes following checks:
            1. If `pass_vars` is defined, `on_end` must be defined.

        Args:
            N/A
        Returns:
            None
        Raises:
            RuntimeError: Raised if `pass_vars` is defined but `on_end`
            callback is not.
        """
        if self.pass_vars and self.on_end is _default_callback:
            raise RuntimeError(
                "If `pass_vars` is defined, `on_end` must be defined.",
            )

    def __call__(self, func):
        """Invoked on decorator instance call.
        Holds logic of the callback process, including invoking callbacks
        and passed function.

        Functions:
            wrapper(*func_args, **func_kwargs):
                Decorator class wrapper accepting `args` and `kwargs`
                for the decorated function. Contains callback logic.

        Args:
            func (Callable): Decorated function to be executed.
        Returns:
            None
        Raises:
            N/A
        """

        def _wrapper(*func_args, **func_kwargs):
            """Decorator class wrapper accepting `args` and `kwargs` for the
            decorated function.
            Calling main method containing callback logic."""
            return self.main(func, func_args, func_kwargs)

        return _wrapper

    def main(self, func, func_args, func_kwargs):
        """Contains callback and execution logic.

        Process:
            1. Validation of decorator arguments.
            2. Setting up custom tracer.
            3. Executing `on_call` callback (if defined).
            4. Executing decorated function.
            5. On success:
                6a. Extracting decorated function local variables.
                7a. Executing `on_success` callback (if defined).
                8a. Returning decorated function result
            5. On error (catching exception from `exception_classes`):
                6b. Extracting decorated function local variables.
                7b. Executing `on_failure` callback (if defined).
                8b. Returning `default_return` value
            9. Reverting to default tracer.
            10. Re-raising decorated function exception if `on_end` callback
                is not defined.
            11. Executing `on_end` callback (if defined).

        Args:
            func(Callable): Decorated function to be executed.
            func_args(tuple): Arguments for the decorated function.
            func_kwargs(dict): Keyword arguments for the decorated function.
        Returns:
            Any: Decorated function result or `default_return` value.
        Raises:
            func_exception: Raised if error occurs during function execution,
            only if `on_end` handler is not defined.
        """
        self.validate_arguments()
        self.set_tracer_profile(self.tracer)
        all_func_kwargs = args_to_kwargs(func, func_args, func_kwargs)
        func_exception, func_result, func_scope_vars = None, None, []
        try:
            self.run_on_call_func(func, all_func_kwargs)
            func_result = func(**all_func_kwargs)
            func_scope_vars = self.get_func_scope_vars()
            self.run_on_success_func(func, func_result, all_func_kwargs)
            return func_result
        except self.exception_classes as ex:
            func_exception = ex
            func_scope_vars = self.get_func_scope_vars()
            self.run_on_failure_func(func, func_exception, all_func_kwargs)
            return self.default_return
        finally:
            self.set_tracer_profile(None)
            if self.on_end is _default_callback and func_exception:
                raise func_exception
            result = func_result if not func_exception else self.default_return
            self.run_on_end_func(
                func, result, func_exception, all_func_kwargs, func_scope_vars
            )



"""Module containing tests for CallPyBack implementation
"""
import pytest
from unittest.mock import MagicMock, call

from callpyback.callpyback import CallPyBack, _default_callback


class Test___init__:
    """Class containing tests for __init__ method"""

    def test_constructor_defaults(self):
        """Tests that constructor sets default values correctly."""
        # Mocks
        # Calls
        decorator = CallPyBack()
        # Assertions
        assert decorator
        assert isinstance(decorator, CallPyBack)
        assert decorator.on_call is _default_callback
        assert decorator.on_success is _default_callback
        assert decorator.on_failure is _default_callback
        assert decorator.on_end is _default_callback
        assert decorator.default_return is None
        assert decorator.pass_vars is None
        assert decorator.exception_classes == (Exception,)
        assert decorator.local_vars == {}
        assert decorator.callbacks == (
            _default_callback,
            _default_callback,
            _default_callback,
            _default_callback,
        )

    def test_constructor_supplied_values(self):
        """Tests that constructor sets supplied values correctly."""
        # Mocks
        on_call = MagicMock()
        on_success = MagicMock()
        on_failure = MagicMock()
        on_end = MagicMock()
        default_return = "default_return"
        pass_vars = ["var1", "var2"]
        exception_classes = (KeyError, ValueError)
        # Calls
        decorator = CallPyBack(
            on_call=on_call,
            on_success=on_success,
            on_failure=on_failure,
            on_end=on_end,
            default_return=default_return,
            pass_vars=pass_vars,
            exception_classes=exception_classes,
        )
        # Assertions
        assert decorator
        assert isinstance(decorator, CallPyBack)
        assert decorator.on_call == on_call
        assert decorator.on_success == on_success
        assert decorator.on_failure == on_failure
        assert decorator.on_end == on_end
        assert decorator.default_return == "default_return"
        assert decorator.pass_vars == ["var1", "var2"]
        assert decorator.exception_classes == (KeyError, ValueError)
        assert decorator.local_vars == {}
        assert decorator.callbacks == (
            on_call,
            on_success,
            on_failure,
            on_end,
        )
        on_call.assert_not_called()
        on_success.assert_not_called()
        on_failure.assert_not_called()
        on_end.assert_not_called()


def create_callpyback_obj(
    on_call=MagicMock(),
    on_success=MagicMock(),
    on_failure=MagicMock(),
    on_end=MagicMock(),
    default_return=None,
    pass_vars=None,
    exception_classes=(Exception,),
):
    """Construct instance of CallPyBack object from supplied or
    default values."""
    instance = CallPyBack(
        on_call=on_call,
        on_success=on_success,
        on_failure=on_failure,
        on_end=on_end,
        default_return=default_return,
        pass_vars=pass_vars,
        exception_classes=exception_classes,
    )
    return instance


class Test_validate_arguments:
    """Class containing tests for validate_arguments method"""

    def test_basic(self):
        """Tests that validate_arguments method calls validation methods."""
        # Mocks
        callpyback_obj = create_callpyback_obj()
        callpyback_obj.validate_callbacks = MagicMock()
        callpyback_obj.validate_pass_vars = MagicMock()
        callpyback_obj.validate_exception_classes = MagicMock()
        # Calls
        callpyback_obj.validate_arguments()
        # Assertions
        callpyback_obj.validate_callbacks.assert_called_once()
        callpyback_obj.validate_pass_vars.assert_called_once()
        callpyback_obj.validate_exception_classes.assert_called_once()


class Test_validate_across_mixins:
    """Class containing tests for validate_across_mixins method"""

    def test_on_end_not_specified_error(self):
        """Tests that error is raised when `on_end` is not specified,
        while `pass_vars` are specified."""
        # Mocks
        callpyback_obj = create_callpyback_obj(
            on_end=_default_callback, pass_vars=("var1", "var2")
        )
        error_msg = "If `pass_vars` is defined, `on_end` must be defined."
        # Calls
        # Assertions
        with pytest.raises(RuntimeError, match=error_msg):
            callpyback_obj.validate_across_mixins()


class Test___call__:
    """Class containing tests for __call__ method"""

    def test_basic(self):
        """Tests that no errors are raised when `__call__` method is called,
        and that `main` method is executed."""
        # Mocks
        callpyback_obj = create_callpyback_obj()
        callpyback_obj.main = MagicMock()
        func = MagicMock()
        args = ()
        kwargs = {}
        # Calls
        result = callpyback_obj.__call__(func, *args, **kwargs)()
        # Assertions
        assert result
        assert callable(result)
        callpyback_obj.main.assert_called_once_with(func, args, kwargs)


class Test_main:
    """Class containing tests for main method"""

    def test_success(self):
        """Tests that `main` method is executed successfully, with correct
        steps without raising any exceptions."""
        # Mocks
        func = MagicMock(return_value="func_return")
        args = (1,)
        kwargs = {"var": "value"}
        all_kwargs = {"args": 1, "var": "value"}
        callpyback_obj = create_callpyback_obj()
        callpyback_obj.tracer = MagicMock()
        callpyback_obj.validate_arguments = MagicMock()
        callpyback_obj.set_tracer_profile = MagicMock()
        callpyback_obj.run_on_call_func = MagicMock()
        callpyback_obj.get_func_scope_vars = MagicMock(return_value=[])
        callpyback_obj.run_on_success_func = MagicMock()
        callpyback_obj.run_on_failure_func = MagicMock()
        callpyback_obj.run_on_end_func = MagicMock()
        # Calls
        result = callpyback_obj.main(func, args, kwargs)
        # Assertions
        assert result == "func_return"
        callpyback_obj.validate_arguments.assert_called_once()
        callpyback_obj.set_tracer_profile.assert_has_calls(
            [call(callpyback_obj.tracer), call(None)]
        )
        callpyback_obj.run_on_call_func.assert_called_with(
            func,
            all_kwargs,
        )
        func.assert_called_once_with(**all_kwargs)
        callpyback_obj.get_func_scope_vars.assert_called_once()
        callpyback_obj.run_on_success_func.assert_called_once_with(
            func,
            result,
            all_kwargs,
        )
        callpyback_obj.run_on_failure_func.assert_not_called()
        callpyback_obj.run_on_end_func.assert_called_once_with(
            func, result, None, all_kwargs, []
        )

    def test_failure_unhandled_ex(self):
        """Tests that `main` method raises an exception when exception is
        raised in decorated function when `on_end` callback not specified."""
        # Mocks
        func = MagicMock(side_effect=Exception("some error"))
        args = (1,)
        kwargs = {"var": "value"}
        all_kwargs = {"args": 1, "var": "value"}
        callpyback_obj = create_callpyback_obj(on_end=_default_callback)
        callpyback_obj.tracer = MagicMock()
        callpyback_obj.validate_arguments = MagicMock()
        callpyback_obj.set_tracer_profile = MagicMock()
        callpyback_obj.run_on_call_func = MagicMock()
        callpyback_obj.get_func_scope_vars = MagicMock(return_value=[])
        callpyback_obj.run_on_success_func = MagicMock()
        callpyback_obj.run_on_failure_func = MagicMock()
        callpyback_obj.run_on_end_func = MagicMock()
        # Calls
        with pytest.raises(Exception, match="some error"):
            _ = callpyback_obj.main(func, args, kwargs)
        # Assertions
        callpyback_obj.validate_arguments.assert_called_once()
        callpyback_obj.set_tracer_profile.assert_has_calls(
            [call(callpyback_obj.tracer), call(None)]
        )
        callpyback_obj.run_on_call_func.assert_called_with(
            func,
            all_kwargs,
        )
        func.assert_called_once_with(**all_kwargs)
        callpyback_obj.get_func_scope_vars.assert_called_once()
        callpyback_obj.run_on_success_func.assert_not_called()
        callpyback_obj.run_on_failure_func.assert_called_once()
        callpyback_obj.run_on_end_func.assert_not_called()

    def test_failure_handled_ex(self):
        """Tests that `main` method does not raise an exception when exception
        is raised in decorated function when on_end callback is specified."""
        # Mocks
        func = MagicMock(side_effect=Exception("some error"))
        args = (1,)
        kwargs = {"var": "value"}
        all_kwargs = {"args": 1, "var": "value"}
        callpyback_obj = create_callpyback_obj()
        callpyback_obj.tracer = MagicMock()
        callpyback_obj.validate_arguments = MagicMock()
        callpyback_obj.set_tracer_profile = MagicMock()
        callpyback_obj.run_on_call_func = MagicMock()
        callpyback_obj.get_func_scope_vars = MagicMock(return_value=[])
        callpyback_obj.run_on_success_func = MagicMock()
        callpyback_obj.run_on_failure_func = MagicMock()
        callpyback_obj.run_on_end_func = MagicMock()
        # Calls
        _ = callpyback_obj.main(func, args, kwargs)
        # Assertions
        callpyback_obj.validate_arguments.assert_called_once()
        callpyback_obj.set_tracer_profile.assert_has_calls(
            [call(callpyback_obj.tracer), call(None)]
        )
        callpyback_obj.run_on_call_func.assert_called_with(
            func,
            all_kwargs,
        )
        func.assert_called_once_with(**all_kwargs)
        callpyback_obj.get_func_scope_vars.assert_called_once()
        callpyback_obj.run_on_success_func.assert_not_called()
        callpyback_obj.run_on_failure_func.assert_called_once()
        callpyback_obj.run_on_end_func.assert_called_once()



"""Module containing BaseCallBackMixin implementation"""

import inspect
import threading

from callpyback.utils import _default_callback

CALLBACK_LABELS = ("on_call", "on_success", "on_failure", "on_end")
CALLBACK_PARAMETER_MAP = {
    "on_call": {"func", "func_kwargs"},
    "on_success": {"func", "func_result", "func_kwargs"},
    "on_failure": {"func", "func_exception", "func_kwargs"},
    "on_end": {
        "func",
        "func_result",
        "func_exception",
        "func_kwargs",
        "func_scope_vars",
    },
}


class BaseCallBackMixin:
    """Class implementing basic callback features.

    Attributes:
        N/A

    Methods:
        validate_callbacks():
            Validates callbacks passed to class constructor.
        run_callback_func(func, func_kwargs):
            Executes given callback function with given kwargs.
        run_on_call_func(func, func_args, func_kwargs):
            Generates kwargs for given `on_call` callback function and executes
            it with generated kwargs.
        run_on_success_func(func, func_result, func_args, func_kwargs):
            Generates kwargs for given `on_success` callback function and
            executes it with generated kwargs.
        run_on_failure_func(func, func_exception, func_args, func_kwargs):
            Generates kwargs for given `on_failure` callback function and
            executes it with generated kwargs.
        run_on_end_func(func, func_result, func_exception, func_args,
        func_kwargs, func_scope_vars):
            Generates kwargs for given `on_end` callback function and executes
            it with generated kwargs.
        get_callback_kwargs(callback, **kwargs):
            Generates kwargs for given callback function.

    """

    def __init__(
        self,
        on_call=_default_callback,
        on_success=_default_callback,
        on_failure=_default_callback,
        on_end=_default_callback,
        **kwargs,
    ):
        """Class constructor. Sets instance variables.

        Args:
            on_call (Callable, optional): Called before execution.
                Defaults to DEFAULT_ON_CALL_LAMBDA.
            on_success (Callable, optional): Called after success.
                Defaults to DEFAULT_ON_SUCCESS_LAMBDA.
            on_failure (Callable, optional): Called after failed execution.
                Defaults to DEFAULT_ON_FAILURE_LAMBDA.
            on_end (Callable, optional): Called after execution.
                Defaults to DEFAULT_ON_END_LAMBDA.

        Returns:
            BaseCallBackMixin: mixin instance
        Raises:
            N/A
        """
        super().__init__(**kwargs)
        self.on_call = on_call
        self.on_success = on_success
        self.on_failure = on_failure
        self.on_end = on_end
        self.callbacks = (
            self.on_call,
            self.on_success,
            self.on_failure,
            self.on_end,
        )

    def validate_callbacks(self):
        """Validates callbacks passed to class constructor.

        Executes following checks:
            1. Callback must be a Callable type.
            2. Callback cannot be an async coroutine.
            3. Callback must accepted some or none of the parameters specified
               in CALLBACK_PARAMETER_MAP.

        Args:
            N/A
        Returns:
            None
        Raises:
            TypeError: Raised if one of the callbacks is not Callable.
            TypeError: Raised if one of the callbacks is an async coroutine.
            AssertionError: Raised if any callbacks haves invalid parameters.
        """
        for callback_name, callback in zip(CALLBACK_LABELS, self.callbacks):
            if not callable(callback):
                raise TypeError("Callback must be a callable.")
            if inspect.iscoroutinefunction(callback):
                raise TypeError(
                    f"Callback `{callback.__name__}` cannot be a coroutine."
                )
            self.validate_callback_parameters(callback, callback_name)

    def validate_callback_parameters(self, callback, callback_name):
        """Validates callback parameters to be accepted by the callback.

        Args:
            callback (func): callback function.
            callback_name (str): name of the callback function.
        Returns:
            None
        Raises:
            AssertionError: Raised if some callback parameters are invalid.
        """
        found_params = sorted(inspect.signature(callback).parameters)
        expected_params = sorted(CALLBACK_PARAMETER_MAP[callback_name])
        if not set(found_params).issubset(set(expected_params)):
            raise AssertionError(
                f"Signature of callback `{callback_name}` is invalid.\n"
                f"Expected: No parameter or combination of: "
                f"{','.join(expected_params)}.\n"
                f"Found: {','.join(found_params)}."
            )

    def run_callback_func(self, func, func_kwargs):
        """Executes given callback function with given kwargs.
        If callback function was decorated with a `@backgroung_callback`,
        function is executed on a new thread in the background.

        Args:
            func (Callable): Callback function to be executed.
            func_kwargs (dict): Keyword arguments to be passed to the callback.
        Returns:
            None
        Raises:
            N/A
        """
        if hasattr(func, "__background_callpyback__"):
            task = threading.Thread(target=func, args=(), kwargs=func_kwargs)
            task.start()
        else:
            func(**func_kwargs)

    def run_on_call_func(self, func, func_kwargs):
        """Generates kwargs for given `on_call` callback function and executes
        it with generated kwargs.

        Args:
            func (func): decorated function
            func_kwargs (dict): Keyword arguments for the decorated function.
        Returns:
            None
        Raises:
            N/A
        """
        on_call_kwargs = self.get_callback_kwargs(
            callback=self.on_call, func=func, func_kwargs=func_kwargs
        )
        self.run_callback_func(self.on_call, on_call_kwargs)

    def run_on_success_func(self, func, func_result, func_kwargs):
        """Generates kwargs for given `on_success` callback function and
        executes it with generated kwargs.

        Args:
            func (func): decorated function
            func_result (Any): Result of the decorated function.
            func_kwargs (dict): Keyword arguments for the decorated function.
        Returns:
            None
        Raises:
            N/A
        """
        on_success_kwargs = self.get_callback_kwargs(
            callback=self.on_success,
            func=func,
            func_result=func_result,
            func_kwargs=func_kwargs,
        )
        self.run_callback_func(self.on_success, on_success_kwargs)

    def run_on_failure_func(self, func, func_exception, func_kwargs):
        """Generates kwargs for given `on_failure` callback function and
        executes it with generated kwargs.

        Args:
            func (func): decorated function
            func_exception (Exception): Exception raised by decorated function.
            func_kwargs (dict): Keyword arguments for the decorated function.
        Returns:
            None
        Raises:
            N/A
        """
        on_failure_kwargs = self.get_callback_kwargs(
            callback=self.on_failure,
            func=func,
            func_exception=func_exception,
            func_kwargs=func_kwargs,
        )
        self.run_callback_func(self.on_failure, on_failure_kwargs)

    def run_on_end_func(
        self, func, func_result, func_exception, func_kwargs, func_scope_vars
    ):
        """Generates kwargs for given `on_end` callback function and executes
        it with generated kwargs.

        Args:
            func (func): decorated function
            func_result (Any): Result of the decorated function.
            func_exception (Exception): Exception raised by decorated function.
            func_kwargs (dict): Keyword arguments for the decorated function.
            func_scope_vars (dict): Scope variables for the decorated function.
        Returns:
            None
        Raises:
            N/A
        """
        on_end_kwargs = self.get_callback_kwargs(
            callback=self.on_end,
            func=func,
            func_result=func_result,
            func_exception=func_exception,
            func_kwargs=func_kwargs,
            func_scope_vars=func_scope_vars,
        )
        self.run_callback_func(self.on_end, on_end_kwargs)

    def get_callback_kwargs(self, callback, **kwargs):
        """Generates kwargs for given callback function.
        Only kwargs specified in callback signature will be passed to callback.
        This is to allow omitting unused parameters in user created callbacks.

        Args:
            kwargs (dict): Decorated function keyword arguments.

        Returns:
            dict: Callback function keyword arguments.
        """
        callback_kwargs = {}
        params = inspect.signature(callback).parameters
        for kwarg_name, kwarg_value in kwargs.items():
            if kwarg_name in params:
                callback_kwargs[kwarg_name] = kwarg_value
        return callback_kwargs



from setuptools import setup, find_packages
from pathlib import Path

CUR_DIR = Path(__file__).parent
LONG_DESCRIPTION = (CUR_DIR / "README.md").read_text()

setup_args = dict(
    name="callpyback",
    packages=find_packages(),
    version="v1.2.0",
    description="Simple and readable Pure Python callbacks!",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="samuelgregorovic",
    author_email="samuelgregorovic@gmail.com",
    url="https://github.com/samuelgregorovic/callpyback",
    download_url="https://github.com/samuelgregorovic/callpyback/archive/refs/tags/v.1.2.0.tar.gz",
    keywords=["callpyback", "callback", "python", "pure", "pythonic", "background"],
    classifiers=[],
)

install_requires = []

if __name__ == "__main__":
    setup(**setup_args, install_requires=install_requires)

# $ python setup.py sdist
# $ twine upload dist/callpyback-v1.0.0.tar.gz
